(** Time-stamp: <modified the 19/05/2011 (at 11:28) by Erwan Jahier> *)


open Lxm
open Errors
open SyntaxTree
open SyntaxTreeCore

(* get the first package in the package/model list *)
let rec first_pack_in =
  function
    | (NSPack pi)::_ -> pi.it.pa_name
    | (NSModel _)::tail -> first_pack_in tail
    | [] -> raise (Global_error "No package has been provided")

let (doit : SyntaxTree.pack_or_model list -> Ident.idref option -> unit) = 
  fun srclist main_node ->

    let syntax_tab = SyntaxTab.create srclist in
      (* Pour chaque package, on a un solveur de rfrences
         globales, pour les types, const et node :
         - les rfrences pointes (p::n) sont recherches
         directement dans la syntax_tab puisqu'il n'y a pas 
         d'ambiguit
         - les rfrences simples sont recherches :
         . dans le pack lui-mme
         . dans un des packs dclars "uses", avec
         priorit dans l'ordre
      *)
    let lzcomp = LazyCompiler.create syntax_tab in
    Verbose.exe ~level:2 (fun () -> SyntaxTab.dump syntax_tab);
    Ident.set_dft_pack_name (first_pack_in srclist);

    let zelic = match main_node with
    | None -> LazyCompiler.compile_all lzcomp
    | Some main_node -> 
       if !Global.compile_all_items then
          LazyCompiler.compile_all lzcomp
       else
          LazyCompiler.compile_node lzcomp main_node
   in
   (* alias des types array *)
   let zelic = DoAliasTypes.doit zelic in

   (* limination polymorphisme  surcharge *)
   let zelic = DoNoPoly.doit zelic in

   LicPrg.to_file !Global.oc zelic
