
(*
Source 2 source transformation :
élimine polymorphisme et surchare

CONDITION :
- il est préférable d'appeler
  ce module AVANT DoAliasTypes,
  sinon on risque d'avoir des alias bizare, du style :
  'anynum_4_7_int'
  au lieu de
  'int_4_7'
  Mais bon, normalement c'est quand même correct ...
*)

open Lxm
open Eff

let dbg=Verbose.get_flag "poly"


let rec doit (inp : LicPrg.t) : LicPrg.t =
   (* n.b. on fait un minumum d'effet de bord pour
      pas avoir trop d'acummulateur ... *)
   let res = ref inp in
   (* on gère une table des types alias pour limiter
      leur création, initialisée par le parcours des
      types *)
   let abstract_types = Hashtbl.create 50 in

   (** TRAITE LES TYPES *)
   let do_type k (te:Eff.type_) =
      ()
   in
   LicPrg.iter_types do_type inp;

   (** TRAITE LES CONSTANTES *)
   let do_const k (ec: Eff.const) =
      ()
   in 
   LicPrg.iter_consts do_const inp ;

   (** TRAITE LES NOEUDS *)
   let rec do_node k (ne:Eff.node_exp) =
      if Eff.node_is_poly ne then
         Verbose.printf ~flag:dbg "#DBG: (No Poly) node %s is POLY\n"
            (Eff.string_of_node_key ne.node_key_eff)
      else
         Verbose.printf ~flag:dbg "#DBG: (No Poly) node %s is NOT poly\n"
            (Eff.string_of_node_key ne.node_key_eff);
         let def' = match ne.def_eff with
         | ExternEff -> ne.def_eff 
         | AbstractEff _ -> assert false
         | BodyEff nb ->
            BodyEff (do_body None None nb)
         in
         ()

   (** TRAITEMENT DES BODY *)
   and do_body
      (a_is: Eff.type_ option)
      (an_is: Eff.type_ option)
      (nb: Eff.node_body) : Eff.node_body =
      (* parcours les expressions du body
         à la recherche d'appel ne noeuds poly *)
      let rec do_exp e =
         e
      and do_assert a =
         Lxm.flagit (
            do_exp a.it
         ) a.src
      and do_eq eq =
         Lxm.flagit (
            fst eq.it,
            do_exp (snd eq.it)
         ) eq.src
      in
      {
         asserts_eff = List.map do_assert nb.asserts_eff;
         eqs_eff = List.map do_eq nb.eqs_eff;
      }

   (* TRAITEMENT DES EXP *)
   and do_exp a_is an_is e =
      let typ' = List.map (do_type a_is an_is) e.ve_typ in
      let core' = match e.ve_core with
      | CallByPosEff (posop, OperEff ops) -> (
         let ops' = OperEff (List.map (do_exp a_is an_is) ops) in
         match posop.it with
         | Predef (pop,sas) ->
            (* pop : Predef.op *) 
            assert false
         | CALL ne ->
            (* ne: node_exp srcflagged *)
            if Eff.node_is_poly ne.it then
               assert false
            else
               CallByPosEff (posop, ops')
         | x ->
            (* dans tout les autre cas, raf ? *)
            CallByPosEff (flagit x posop.src, ops')
      )
      | CallByNameEff (namop, idops) ->
      assert false
      in
      { e with ve_core = core'; ve_typ = typ' }

   (* TRAITEMENT DES TYPE EFF *)
   and do_type a_is an_is t = match t with
      | TypeVar Any -> (match a_is with Some t -> t | _ -> assert false)
      | TypeVar AnyNum -> (match an_is with Some t -> t | _ -> assert false) 
      | Array_type_eff    (t, sz) -> Array_type_eff (do_type a_is an_is t, sz)
      | Abstract_type_eff (id, t) ->
         (* i.e. alias *)
         let t' = do_type a_is an_is t in
         if t <> t' then
            get_abstract_type id a_is an_is t' 
         else
            Abstract_type_eff (id, t)
      (* Dans tous les autres cas, raf *)
      | Struct_type_eff _
      | Bool_type_eff
      | Int_type_eff
      | Real_type_eff
      | External_type_eff _
      | Enum_type_eff _ -> t

   (* RECHERCHE/CREATION DE TYPE ALIAS *)
   and get_abstract_type old_id a_is an_is teff =
      (* old_id est l'ancien nom, teff la nouvelle def
         normallement, exactement a_is OU BIEN an_is
         vaut qq chose *)
      try Hashtbl.find abstract_types teff
      with Not_found ->
         (* nom sympa ... ? *)
         let sfx = match a_is, an_is with
            | Some t, None
            | None, Some t -> Eff.ident_of_type t
            | _ -> assert false
         in
         (* pour garantif l'unicité *)
         let tid = LicPrg.fresh_type_id !res (fst old_id) ((snd old_id)^(snd sfx)) in
         let ref_teff = Abstract_type_eff (tid, teff) in
         res := LicPrg.add_type tid ref_teff !res;
         Hashtbl.add abstract_types teff ref_teff;
         ref_teff 
   in
   (* LET's GO *)
   LicPrg.iter_nodes do_node inp;
   !res
