
(* Réorganisation de la compil :
    Un LicPrg est :
    - un programme Lustre SIMPLE, vérifié et cohérent,
      à base de Eff.xx
    - du lv6, on passe à LicPrg via LazyCompiler, qui fait
      UNIQUEMENT le boulot de base :
      * dé-packaging
      * résolution de l'ordre sup statique, y compris la
        récursion, en tirant "le fil" du main node
		* SAUF pour les macros prédéfinies (non-programmables)
		* résolution de la surcharge 
	 - les transformations (expansions etc.) qui étaient faites
      dans LazyCompiler sont (appelées à devenir) des phases 
      apr coup du type LicPrg -> LicPrg
*)
(*
	Pas très différent des infos de LazyCompiler.t
	Sauf que on utilise des map
	
*)

module ItemKeyMap = struct
	include Map.Make (
		struct
			type t = Eff.item_key
			let compare = compare
		end
	)
	let dummy () = "dummy item: add things below to complete the module"
end

module NodeKeyMap = struct
	include Map.Make (
		struct
			type t = Eff.node_key
			let compare = compare
		end
	)
	let dummy () = "dummy item: add things below to complete the module"
end

type t = {
  types  : Eff.type_ ItemKeyMap.t;
  consts : Eff.const ItemKeyMap.t;
  nodes  : Eff.node_exp NodeKeyMap.t;
  (* consts : (Eff.item_key, Eff.const    Eff.check_flag) Hashtbl.t; *)
  (* nodes  : (Eff.node_key, Eff.node_exp Eff.check_flag) Hashtbl.t; *)
}

let empty = {
   types = ItemKeyMap.empty;
   consts = ItemKeyMap.empty;
   nodes = NodeKeyMap.empty
}

(* KoKêterie : pour changer ds suffixes 
   numériques, produit :
   a, b, c, d, e, aa, ab, ac, ad, ae, etc ...
Pour tester :
for i = 0 to 42 do
Printf.printf "%3d -> %s\n" i (pretty_sfx i)
done
*)
let rec pretty_sfx i =
   if i = 0 then ""
   else
      (pretty_sfx ((i-1)/5))^(Char.escaped (char_of_int (97 + (i-1) mod 5)))

(** CREER DES IDENTS TOUT FRAIS *)
let fresh_type_id this pname pfx =
   let rec fresh x =
      let id = Printf.sprintf "%s%s" pfx (pretty_sfx x) in
      let res = Ident.make_long pname id in
      if ItemKeyMap.mem res this.types then fresh (x+1)
      else res
   in
   fresh 0

(** RECHERCHE *)
let find_type  this k = ItemKeyMap.find k this.types
let find_const this k = ItemKeyMap.find k this.consts
let find_node  this k = NodeKeyMap.find k this.nodes
   

(** PARCOURS *)
let fold_consts (f: Eff.item_key -> Eff.const -> 'a -> 'a) (this:t) (accin:'a) : 'a =
   ItemKeyMap.fold f this.consts accin
let fold_types (f: Eff.item_key -> Eff.type_ -> 'a -> 'a) (this:t) (accin:'a) : 'a =
   ItemKeyMap.fold f this.types accin
let fold_nodes (f: Eff.node_key -> Eff.node_exp -> 'a -> 'a) (this:t) (accin:'a) : 'a =
   NodeKeyMap.fold f this.nodes accin

let iter_consts (f: Eff.item_key -> Eff.const -> unit) (this:t) : unit =
   ItemKeyMap.iter f this.consts
let iter_types (f: Eff.item_key -> Eff.type_ -> unit) (this:t) : unit = 
   ItemKeyMap.iter f this.types
let iter_nodes (f: Eff.node_key -> Eff.node_exp -> unit) (this:t) : unit =
   NodeKeyMap.iter f this.nodes



let add_type (k:Eff.item_key) (v:Eff.type_) (prg:t) : t =
   { prg with types = ItemKeyMap.add k v prg.types }

let add_const (k:Eff.item_key) (v:Eff.const) (prg:t) : t =
   { prg with consts = ItemKeyMap.add k v prg.consts }

let add_node (k:Eff.node_key) (v:Eff.node_exp) (prg:t) : t =
Verbose.printf ~level:3 "## LicPrg.add_node %s\n" (LicDump.string_of_node_key_rec k);
   { prg with nodes = NodeKeyMap.add k v prg.nodes }


let dump_entete oc = 
  let time = Unix.localtime (Unix.time ()) in
  let sys_call, _ = Array.fold_left 
    (fun (acc,i) x -> 
       if 70 < i + (String.length x) then 
	 acc ^ "\n--\t\t" ^ x, String.length ("\n--\t\t" ^ x)
       else 
	 acc ^ " " ^ x , (i+1+(String.length x))
    )
    ("",0) 
    Sys.argv
   and
      date = Printf.sprintf "%d/%02d/%02d"
         (1900+time.Unix.tm_year)
         (time.Unix.tm_mon+1)
         (time.Unix.tm_mday)
   and
      time_str = Printf.sprintf "%02d:%02d:%02d"
         (time.Unix.tm_hour)
         (time.Unix.tm_min)
         (time.Unix.tm_sec) 
   (*   and user = Unix.getlogin () *)
   and hostname = Unix.gethostname ()
   in
      (* Printf.fprintf oc "-- lus2lic version %s\n" Version.str; *)
      (* Printf.fprintf oc "-- cmd: %s\n" sys_call; *)
      (* Printf.fprintf oc "-- host: %s date: %s time: %s\n" hostname date time_str  *)
      Printf.fprintf oc "-- This file was generated by lus2lic version %s.\n" Version.str;
      Printf.fprintf oc "-- %s\n" sys_call;
      Printf.fprintf oc "-- on %s the %s at %s\n" hostname date time_str 

let to_file (oc: out_channel) (this:t) = 
   dump_entete oc;
   (* On imprime dans l'ordre du iter, donc pas terrible ???
   *)
   ItemKeyMap.iter
      (fun tn te -> output_string !Global.oc (LicDump.type_decl tn te)) this.types;
   ItemKeyMap.iter
      (fun cn ce -> output_string !Global.oc (LicDump.const_decl cn ce)) this.consts;

   (* Pour les noeuds, pas sur que ça marche tant qu'on n'a
      pas séparés les transformations  source_to_source du LazyCompiler :
      en cas d'expansion, il y avait cette remarque :
           nb: we print res_struct, but do not return it from
           node_check, because the structure and array expansion
           modify (instanciate) the node profiles.  *)

   NodeKeyMap.iter
      (fun _ nexp -> output_string !Global.oc (LicDump.node_of_node_exp_eff nexp)) this.nodes

